program ValidatorExample;
  { (c) 2000, T.Verhoeff @ TUE.NL }
  { Example relaxed input validator that checks multiple files }

uses
  Dos,
  RobIn; { Robust Input }

const
  MdfDir    = 'd:\tom\ioi2000\robin\'; { directory with general message
                                         definitions }
  NCases    = 5; { number of test cases }

type
  CaseNumber = 1 .. NCases;

  { Format definition:

    The first line of file VALIDEX.IN contains a string S of
    at least 3 and at most 20 lower case characters or digits,
    not starting with a digit.
    The second line contains two integers N and K,
    with 1 <= K <= N <= 100.
    The remaining N lines each contain K upper case characters,
    without blanks between them.
  }

const
  TaskDir   = '';
  TaskName  = 'ValidEx';
  ProgName  = TaskName;
  FmtName   = TaskName + '.in';
  InpName   = TaskName + '.i'; { append two-digit test case number }
  LogName   = TaskName + '.log';
  MdfName   = TaskName + '.mdf';

  { bounds }
  MinLenS   =   3;
  MaxLenS   =  20;
  MinN      =   1;
  MaxN      = 100;
  MinK      =   1;
  MaxK      = 100;

var
  { input variables }
  S: String [ MaxLenS ];
  N: MinN .. MaxN;
  K: MinK .. MaxK;
  C: array [ 1 .. MaxN, 1 .. MaxK ] of Char;
     { C[i,j] is j-th character on line k+2 }

  { auxiliary variables }
  i: 1 .. MaxN; { to traverse the N lines }
  j: 1 .. MaxK; { to traverse the K characters on a line }

procedure ReadInput ( var inFile: RobInFile );
  { pre:  inFile is open
    post: S, N, K, C have been read; inFile is still open }
  begin
    { read first line }
    S := ReadStringRIF ( inFile, Contiguous, 'S',
                         UpperCaseChars + DigitChars, MinLenS, MaxLenS )
  ; AssertRIF ( inFile, not ( S [ 1 ] in DigitChars ),
                AssertWarning, 900, S [ 1 ] )
  ; ReadlnRIF ( inFile )

    { read second line }
  ; N := ReadIntegerRIF ( inFile, Separated, 'N', MinN, MaxN )
  ; K := ReadIntegerRIF ( inFile, Separated, 'K', MinK, MaxK )
  ; ReadlnRIF ( inFile )

    { read remaining N lines }
  ; for i := 1 to N do begin
      for j := 1 to K do begin
        C[i,j] := ReadCharRIF ( inFile, Contiguous,
                                'C[' + IntToStr ( i, 1 ) + ',' +
                                  IntToStr ( j, 1 ) + ']',
                                LowerCaseChars )
      end { for j }
    ; ReadlnRIF ( inFile )
    end { for i }

  end; { ReadInput }

procedure DoFile ( inPath: String );
  begin
    OpenRIF ( inFile, FmtName, inPath, RelaxedChecking )

  ; ReadInput ( inFile )

  ; CloseRIF ( inFile )

  ; if IsAbortedRIF ( inFile ) then begin
      writeln ( 'Input file "', inPath, '" is NOT valid' )
    end { then }
    else begin
      writeln ( 'Input file "', inPath, '" appears valid' )
    end { if }

    { possibly do some further analysis of the input variables here ... }

  end; { DoFile }


{ ----- there should be no need to touch framework below this line ----- }
{ The framework calls DoFile ( inFilePath ) for each file. }

procedure HandleCommandLine;
  var
    i   : Integer;   { to traverse the command-line parameters }
    s   : SearchRec; { to traverse the directory in case of wildcards }
    dir : DirStr;    { to capture result from FSplit }
    name: NameStr;   { to capture result from FSplit }
    ext : ExtStr;    { to capture result from FSplit }
  begin

    for i := 1 to ParamCount do begin
      FSplit ( ParamStr ( i ), dir, name, ext )
    ; FindFirst(ParamStr(i), 0, s)
    ; while DosError = 0 do begin
        DoFile ( dir + s.Name )
      ; FindNext ( s )
      end { while }
    ; if DosError <> 18 then begin
        writeln ( ParamStr ( i ), ' not found (', DosError, ')' )
      end { if }
    end { for i }

  end; { HandleCommandLine }

function FillLeadingZeroes ( s: String; width: integer ): String;
  { pre: width > 0
    ret: s with minimal # leading zeroes to make length >= width }
  begin

    while Length ( s ) < width do begin
      s := '0' + s
    end { while }

  ; FillLeadingZeroes := s
  end; { FillLeadingZeroes }

procedure HandleTaskDirectory;
  var
    testCase: CaseNumber; { to traverse the example and test cases }
  begin
    DoFile ( TaskDir + FmtName ) { the example input file }

  ; for testCase := 1 to NCases do begin
       DoFile ( TaskDir + InpName +
                  FillLeadingZeroes ( IntToStr ( testCase, 1 ), 2 ) )
    end { for }

  end; { HandleTaskDirectory }

begin
  writeln ( ProgName, ': Input validator for ', TaskDir + FmtName )

; BeginLogging ( logFile, LogName,
                 OrganizerDiagnostics, {UnlimitedReporting} 2,
                 MdfDir, mdfName )

; if ParamCount = 0 then begin
    { do all examples and test cases in task directory }
    HandleTaskDirectory
  end { then }
  else begin
    { do all files on command line }
    HandleCommandLine
  end { else }

; EndLogging ( LogName <> '' )
end.