unit RobIn;
  { (c) 2000, Tom Verhoeff (TUE) }

  { Unit for robust input from text files with IOI-style format.
    Not intended for interactive text i/o.
    Not optimized for speed.

    For documentation see Doc.txt.
    For examples see Example.pas and ValidEx.pas.

    Needs files TPRTerr.mdf and RobInGen.mdf.
  }

{ Compiler Directives: Safety first, performance last.
  Also recommended for programs that use RobIn! }
{ Where really needed, override the defaults below. }
{$B- Boolean Evaluation: Short Circuit (needed for correctness!) }
{$I+ I/O Checking      : On }
{$Q+ Overflow Checking : On }
{$R+ Range Checking    : On }
{$S+ Stack Checking    : On }

interface {
========= }

const
  RobInCopyright = 'Copyright (c) 2000 T.Verhoeff @ TUE.NL';
  RobInVersion   = '0.3 (beta)';


{ The following constant and type definitions are directly relevant
  to the use of RobIn.  They are needed to specify the interface
  routines and their parameters.  }

type
  DiagnosticCategory    = ( { abbreviation }
    IOError,                { IOE }
    InternalWarning,        { IW }  { currently not used }
    InternalError,          { IE }
    UsageWarning,           { UW }
    UsageError,             { UE }
    MetaFormatWarning,      { MFW }
    MetaFormatError,        { MFE }
    FormatWarning,          { FW }
    FormatError,            { FE }
    AssertWarning,          { AW }
    AssertError,            { AE }
    OpenCloseInfo,          { OCI }
    ItemInfo,               { II }
    WhitespaceInfo,         { WI }
    DebugInfo               { DI }
  ); { every diagnostic message is issued together with a category }
  AssertDiagnostic      = AssertWarning .. AssertError;
    { used in AssertRIF }
  DiagnosticSet         = Set of DiagnosticCategory;
    { used in BeginLogging to filter diagnostic output }
  DiagnosticCount       = array [ DiagnosticCategory, Boolean ] of LongInt;
    { For var diagCount: DiagnosticCount, dc: DiagnosticCategory,
        diagCount [ dc, false ] = # UNreported diagnostics in category dc
        diagCount [ dc, true ]  = # reported diagnostics in category dc }

const
  FirstDiagnostic       = IOError;
  LastDiagnostic        = DebugInfo;
  CompetitorDiagnostics = [ IOError, InternalError,
                            FormatWarning, FormatError,
                            AssertWarning, AssertError,
                            OpenCloseInfo ];
    { normal diagnostics for competitors }
  OrganizerDiagnostics  = [ IOError .. OpenCloseInfo ];
    { normal diagnostics for organizers }
  AllDiagnostics        = [ FirstDiagnostic .. LastDiagnostic ];
  ExtensiveDiagnostics  = AllDiagnostics - [ DebugInfo ];
    { all diagnostics, except DebugInfo }
  AbortReadDiagnostics  = [ IOError,
                            MetaFormatError, FormatError,
                            AssertError ];
    { diagnostics that abort the reading of a robust input file,
      but not the execution of the program }


type
  CheckingLevel = ( RelaxedChecking, StrictChecking );
  { used in OpenRIF,
    RelaxedChecking gives format warnings for item composition mistakes.
    StrictChecking gives format errors for item composition mistakes. }


type
  ItemCategory       = ( NoItem, CharacterItem, IntegerItem, StringItem );
    { ItemCategory is not part of the interface, but is needed for
      RobInFile below }
  ItemComposition    = ( Contiguous, Separated ); { for Read<Item>RIF }


const
  SpecialMessageNr   =   0; { always associated with empty message }
  MinMessageNumber   =   1;
  MaxMessageNumber   = 999;
  MinAssertMsgNr     = 900;
  MaxAssertMsgNr     = MaxMessageNumber;

  NoReporting        =   0;
  UnlimitedReporting = 101;
  MaxMsgCount        = 100; { max. number of suppressed messages counted }

type
  MessageNumber = SpecialMessageNr .. MaxMessageNumber;
  AssertMsgNr   = MinAssertMsgNr .. MaxAssertMsgNr;
    { used in AssertRIF }
  LimitRange    = NoReporting .. UnlimitedReporting;
    { used in BeginLogging }
  MsgCountRange = - UnlimitedReporting .. MaxMsgCount;
  MessageCount  = array [ MessageNumber ] of MsgCountRange;
    { For var msgCount: MessageCount, msgNr: MessageNumber,
        msgCount [ msgNr ] = # suppressed messages with msgNumber,
          that is, # messages raised but beyond the reporting limit,
        if equal MaxMsgCount, then this means "MaxMsgCount or more",
        if negative, then minus the number to go before limiting starts
      Limiting can only be done for the following diagnostic categories:
        FW, MFW, AW, UW }


type
  { Variables of type RobInFile should only be manipulated by
    the routines below: OpenRIF, Read<Item>RIF, ReadlnRIF, CloseRIF,
                        AssertRIF, IsAbortedRIF.
    There is no need to understand the fields in order to use this unit. }
  RobInFile    = record   { private fields: only valid after OpenRIF }
    fmtName    : String;  { identifier for format }
    f          : Text;    { the file being read }
    fPath      : String;  { path for file f }
    fOpened    : Boolean; { whether f is indeed open to fPath }
    chLevel    : CheckingLevel;   { whether checking is strict or relaxed }
    precItemCat: ItemCategory;    { category of preceding item on current
                                    format line }
    itemComp   : ItemComposition; { how preceding and current item are
                                    composed }
    itemCat    : ItemCategory;    { category of current item }
    itemIdent  : String;  { identifier for current item }
    itemValue  : String;  { human readable string for item's value }
    nextChar   : Char;    { look-ahead character, has been read from f }
    lineNr     : LongInt; { line number of nextChar, starts at 1 }
    columnNr   : LongInt; { column number of nextChar, starts at 1 }
    charCount  : LongInt; { # characters read from f, not eof/eoln }
    fmtLineNr  : LongInt; { format line number of current item, starts at 1 }
    itemNr     : LongInt; { current item number on format line, starts at 1 }
    itemCount  : LongInt; { # items processed from format }
    aborted    : Boolean; { whether reading from f has been aborted }
    diagCount  : DiagnosticCount; { message output statistics per category }
    msgCount   : MessageCount;    { message limiting statistics per number }
    rifOpen    : Boolean; { also need to check signature = theSignature }
    signature  : LongInt; { to detect initialization }
    end; { record }
  { var rif: RobInFile is a state descriptor of the reading/checking
    process for the related file rif.f }


type
  CharSet = Set of Char; { for range in ReadCharRIF and ReadStringRIF }

const
  Blank             = ' ';         { = SP = chr ( 32 ), acts as separator }
  FirstPrintingChar = chr (  33 ); { = succ ( SP ) }
  LastPrintingChar  = chr ( 126 ); { = pred ( DEL ) }
  PrintingChars     = [ FirstPrintingChar .. LastPrintingChar ];
    { printable non-blank non-extended ASCII
      N.B. Blank not in PrintingChars }
  DigitChars        = [ '0' .. '9' ];
  UpperCaseChars    = [ 'A' .. 'Z' ];
  LowerCaseChars    = [ 'a' .. 'z' ];
  MaxStringLength   = 255;


var { for convenience, most using programs will need these anyway }
  logFile: Text;      { file for diagnostic output }
  inFile : RobInFile; { N.B. it is also declared here for safety,
                        because RobIn can then initialize it }


procedure BeginLogging ( var logFile: Text;
                         const logPath: String;
                         const diagSelection: DiagnosticSet;
                         reportingLimit: LimitRange;
                         const mdfDir,
                               mdfPath: String );
  { pre:  logFile is text file variable for logging,
          logPath is path for log file (used for opening, if nonempty)
          diagSelection is choice of diagnostic messages to log,
          RIF-messages beyond reportingLimit are suppressed,
          mdfDir is path to directory with TPRTerr.mdf and RobInGen.mdf,
          mdfPath is path to message definition file for AssertRIF
            (message numbers in the range 900..999)
    post: (if logPath <> '' then logFile is opened for writing to logPath),
          subsequent diagnostic output will go to logFile,
          diagnostic output is filtered according to diagSelection,
          message definitions in TPRTerr.mdf and RobInGen.mdf have been
            loaded from directory mdfDir (only once),
          if mdfPath <> '' then message definitions have been loaded
            from mdfPath
    usage note: initially,
                diagnostic output goes to standard output, and
                no message definitions are loaded
    usage note: when opening logFile, any existing file is overwritten
                without warning
    usage note: if DebugInfo in diagSelection, then extra information
                about loaded message definitions is provided
  }
procedure EndLogging ( closeLogFile: Boolean );
  { pre:  logging is begun
    post: the loaded message definitions have been unloaded
          if closeLogFile, then the currently active log file is closed }
procedure OpenRIF ( var rif: RobInFile;
                    const formatName, filePath: String;
                    chkLvl: CheckingLevel );
  { pre:  rif is closed or not initialized,
          formatName is an identier for the task and file
          to which the format applies (typically the file name given in
            the task description; it is used in reporting only),
          filePath is path for existing text file (used for opening
            the file)
    post: rif is open,
          rif.f is open for reading from beginning of file named filePath,
          content of rif.f is checked according to checking level chkLvl,
          I/O Error if file named filePath is not found }
procedure CloseRIF ( var rif: RobInFile );
  { pre:  rif is open
    post: rif.f is closed, rif is closed;
          Meta-Format Error if at begin-of-file or not at end-of-file }
function  ReadCharRIF ( var rif: RobInFile;
                        comp: ItemComposition;
                        const itemId: String;
                        range: CharSet ): Char;
  { pre:  rif is open, [ ] < range <= PrintingChars
    ret:  character read
    post: one character is read,
          item composition is handled according to comp,
          for first item on a line comp is adjusted to Contiguous;
          Format Error if at end-of-file or at end-of-line or out-of-range }
function  ReadIntegerRIF ( var rif: RobInFile;
                           comp: ItemComposition;
                           const itemId: String;
                           lowerBound, upperBound: LongInt ): LongInt;
  { pre:  rif is open, lowerBound <= upperBound
    ret:  integer read
    post: one integer number is read,
          item composition is handled according to comp;
          for first item on a line comp is adjusted to Contiguous;
          Format Error if at end-of-file or at end-of-line or
            preceded by junk or out-of-range }
function  ReadStringRIF ( var rif: RobInFile;
                          comp: ItemComposition;
                          const itemId: String;
                          range: CharSet;
                          minLength, maxLength: Integer ): String;
  { pre:  rif is open, [ ] < range <= PrintingChars,
          1 <= minLength <= maxLength <= MaxStringLength
    ret:  string read
    post: one string of at most MaxStringLength characters is read into s,
          item composition is handled according to comp,
          for first item on a line comp is adjusted to Contiguous;
          Format Error if at end-of-file or at end-of-line or too long or
            containing out-of-range characters }
procedure ReadlnRIF ( var rif: RobInFile );
  { pre:  rif is open
    post: one end-of-line is read;
          Format Error if at end-of-file or not at end-of-line }
procedure AssertRIF ( var rif: RobInFile;
                      condition: Boolean;
                      assertDiag: AssertDiagnostic;
                      msgNr: Integer;
                      const extraMsg: String );
  { pre:  rif is open, 900 <= msgNr <= 999
    post: true;
          does nothing if aborted;
          raises assertDiag if condition does not hold;
          aborts on failure if assertDiag = AssertError }
function  IsAbortedRIF ( var rif: RobInFile ): Boolean;
  { pre:  rif is initialized by OpenRIF (and possibly closed already)
    ret:  whether reading was aborted because of a fatal error }

function  IntToStr ( i: LongInt;
                     width: integer ): String;
  { pre:  0 < w <= 255
    ret:  human readable string representation of i,
          left extended with blanks to obtain desired width
    usage note: especially for constructing itemIdent in Read<Item>RIF
                when item is part of a sequence }


var { global variables, manipulate directly for special cases only }
  { Set by BeginLogging, no need to use directly}
  theLogFileP      : ^ Text; { pointer to the log file for diagnostics }
  theDiagSel       : DiagnosticSet; { the selected diagnostic categories }
  theLimitedDiag   : DiagnosticSet; { diagnostics for which RIF reporting
                                      can be limited per message number }
  theReportingLimit: LimitRange;    { the max. number of RIF-messages
    reported (per message number) before suppression sets in;
    suppresses always when equal to NoReporting,
    never suppresses when equal to UnlimitedReporting }


implementation {
============== }

uses Dos;

{ use $DEFINE TEST to write test output in log file (not so much now)
  remove/insert 'x' after leading '{' below to activate/deactivate }
{x$DEFINE TEST}

{ ----- General Facilities ----- }

const
  ModuleName = 'RobIn';

  CtrlShift = 64;
  TabChar   = chr ( ord ( 'I' ) - CtrlShift ); { Control-I = TAB }
  EolnChar  = chr ( ord ( 'M' ) - CtrlShift ); { Control-M = CR }
  EofChar   = chr ( ord ( 'Z' ) - CtrlShift ); { Control-Z }
  { The choice for the encodings above is motivated by the fact that
    when EolnChar or EofChar occur in a text file, they are recognized
    by the Turbo Pascal Run-Time System as eoln (end-of-line) and
    eof (end-of-file).
  }
  Whitespace     = [ EolnChar, Blank ]; { no TabChar for IOI }
  PrintableChars = PrintingChars + [ Blank ];
  Indent         = '  '; { indentation for lines in diagnostic output
                           after first line is two blanks }
  ColSep         = '  '; { separation between columns is two blanks }
  NumWidth       = 5;    { field width for numbers in statistics
                           to make nice columns }
  MaxMsgNrWidth  = 3;    { max width of a MessageNumber }
  MaxMsgCntWidth = 3;    { max width of a MessageCount }

var
  theSignature: LongInt; { should be initialized sufficiently "random" }

procedure InitSignature ( var s: LongInt );
  { pre:  true
    post: s = value derived from current time }
  var
    dt    : DateTime; { to construct parameter for PackTime }
    dow   : Word;     { to capture unneeded result of GetDate }
    sec100: Word;     { to capture unneeded result of GetTime }
  begin
    with dt do begin
      GetDate ( year, month, day, dow )
    ; GetTime ( hour, min, sec, sec100 )
    end { with dt }
  ; PackTime ( dt, s )
  end; { InitSignature }

procedure WriteIdentification ( var outFile: Text );
  { pre:  outFile is open for writing
    post: identification written to outFile }
  begin
    writeln ( outFile, 'Uses ', ModuleName,
                       '  Version ', RobInVersion,
                       '  ', RobInCopyright )
  ; writeln ( outFile )
  end; { WriteIdentification }

procedure ReportDiagnostic ( const routineName: String;
                             diagCat: DiagnosticCategory;
                             msgNr: MessageNumber;
                             const extraMsg: String ); forward;

function  CharToStr ( c: Char ): String;
  { pre:  true
    ret:  human readable string representation of character }
  begin
    if c in PrintableChars then begin
      if c = '''' then
        CharToStr := '''''''''''' { shows as '''' }
      else
        CharToStr := '''' + c + '''' { shows as 'c' }
    end { then }
    else begin
      case c of
        EolnChar: CharToStr := 'End-of-Line'
      ; EofChar:  CharToStr := 'End-of-File'
        else      CharToStr := 'chr(' + IntToStr ( ord ( c ), 1 ) + ')'
      end { case }
    end { else }
  end; { CharToStr }

function  BoolToStr ( b: Boolean ): String;
  { pre:  true
    ret:  human readable string representation of b }
  begin
    case b of
      False: BoolToStr := 'False'
    ; True : BoolToStr := 'True'
    end { case b }
  end; { BoolToStr }

function  UpCaseString ( s: String ): String;
  { pre:  true
    ret:  s where all characters have been converted to upper case }
  var
    i: Integer; { to traverse s }
  begin

    for i := 1 to Length ( s ) do begin
      s [ i ] := UpCase ( s [ i ] )
    end { for i }

  ; UpCaseString := s
  end; { UpCaseString }

function  Abbreviate ( s: String ): String;
  { pre:  true
    ret:  s where all non-upper case characters have been deleted }
  var
    i: Integer; { to traverse s }
  begin
    i := 1

  ; while i <= Length ( s ) do begin
      if s [ i ] in UpperCaseChars then begin
        Inc ( i )
      end { then }
      else begin
        Delete ( s, i, 1 )
      end { else }
    end { while }

  ; Abbreviate := s
  end; { Abbreviate }

function  TrimString ( s: String ): String;
  { pre:  true
    ret:  s without leading and trailing blanks }
  begin

    while ( s <> '' ) and ( s [ 1 ] = Blank ) do begin
      Delete ( s, 1, 1 )
    end { while }

  ; while ( s <> '' ) and ( s [ Length ( s ) ] = Blank ) do begin
      Delete ( s, Length ( s ), 1 )
    end { while }

  ; TrimString := s
  end; { TrimString }

function  SizeCS ( const cs: CharSet ): Integer;
  { pre:  true
    ret:  number of characters in cs }
  var
    c     : Char;    { to traverse cs }
    result: Integer; { to accumulate the result }
  begin
    result := 0

  ; for c := chr ( 0 ) to chr ( 255 ) do begin
      Inc ( c, ord ( c in cs ) )
    end { for c }

  ; SizeCS := result
  end; { SizeCS }

function  MinCS ( const cs: CharSet ): Char;
  { pre:  cs <> [ ]
    ret:  minimum character in cs }
  { impl: Linear Search for smallest character in cs }
  const
    RoutineName = 'MinCS';
  var
    c: Char; { to traverse cs }
  begin
    if cs = [ ] then begin
      ReportDiagnostic ( RoutineName, InternalError, {M#}256, '' )
    end { if }
  ; c := chr ( 0 )

  ; while not ( c in cs ) do begin
      Inc ( c )
    end { while }

  ; MinCS := c
  end; { MinCS }


function  StrCS ( const cs: CharSet ): String;
  { pre:  true
    ret:  string of printing characters in cs }
  var
    c     : Char;   { to traverse cs }
    result: String; { to accumulate result }
  begin
    result := ''

  ; for c := FirstPrintingChar to LastPrintingChar do begin
      if c in cs then begin
        result := result + c
      end { if }
    end { for c }

  ; StrCS := result
  end; { StrCS }


{ -----
  Abstract Data Type MessageTable
    for maintaining tables of message definitions,
    associating a message number with a message text
  ----- }

{ Some constants and types were (out of necessity) already defined
  in the interface. }

const
  MaxMessageLength =  50; { max. length of message text }

type
  Message       = String [ MaxMessageLength ]; { for text of message }
  MessageTableR = record
    tbl: array [ MessageNumber ] of record ptr: ^ Message end
    end;
  MessageTable  = ^ MessageTableR; { This is the ADT, operations below }

var
  { global variables }
  theMsgTbl   : MessageTable;    { the table with message definitions }
  theDiagCount: DiagnosticCount; { message statistics per log session }

function  GetMessage ( const mt: MessageTable;
                       mn: MessageNumber ): Message;
  { pre:  mt <> nil
    ret:  message associated with mn in mt }
  { impl note: GetMessage is called by ReportDiagnostic.
               GetMessage must not call ReportDiagnostic, so as to avoid
                 *  intertwining of lines in distinct diagnostic reports
                 *  nonterminating mutual recursion
  }
  const
    RoutineName = 'GetMessage';
  begin
    if mt = nil then begin
      { Message table does not exist }
      GetMessage := '<<< message table not available >>>'
    end { then }
    else with mt ^, tbl [ mn ] do begin
      if ptr = nil then begin
        { Message not defined }
      ; GetMessage := '<<< message text not available >>>'
      end { then }
      else { ptr <> nil } begin
        GetMessage := ptr ^
      end { if }
    end { else with mt ^ }
  end; { GetMessage }

const
  MaxDCStringLen = 20; { upper bound on length of StringFromDC value }

function  StringFromDC ( dc: DiagnosticCategory ): String;
  { pre:  true
    ret:  human readable string representation of dc }
  begin
    case dc of
      IOError          : StringFromDC := 'I/O Error'
    ; InternalError    : StringFromDC := 'Internal Error'
    ; InternalWarning  : StringFromDC := 'Internal Warning'
    ; UsageWarning     : StringFromDC := 'Usage Warning'
    ; UsageError       : StringFromDC := 'Usage Error'
    ; MetaFormatWarning: StringFromDC := 'Meta-Format Warning'
    ; MetaFormatError  : StringFromDC := 'Meta-Format Error'
    ; FormatWarning    : StringFromDC := 'Format Warning'
    ; FormatError      : StringFromDC := 'Format Error'
    ; AssertWarning    : StringFromDC := 'Assert Warning'
    ; AssertError      : StringFromDC := 'Assert Error'
    ; OpenCloseInfo    : StringFromDC := 'Open/Close Info'
    ; ItemInfo         : StringFromDC := 'Item Info'
    ; WhitespaceInfo   : StringFromDC := 'Whitespace Info'
    ; DebugInfo        : StringFromDC := 'Debug Info'
    end { case dc }
  end; { StringFromDC }

function  StringFromDS ( const ds: DiagnosticSet ): String;
  { pre:  true
    ret:  human-readable string representing ds }
  var
    dc    : DiagnosticCategory; { to traverse ds }
    result: String;             { to accumulate result }
    sep   : String;             { element separator }
  begin
    result := '['
  ; sep    := ''

  ; for dc := FirstDiagnostic to LastDiagnostic do begin
      if dc in ds then begin
        result := result + sep + Abbreviate ( StringFromDC ( dc ) )
      ; sep := ','
      end { if }
    end { for dc }

  ; StringFromDS := result + ']'
  end; { StringFromDS }

procedure InitDiagCount ( var diagCount: DiagnosticCount );
  { pre:  true
    post: diagCount [ dc, b ] = 0 for all dc, b }
  var
    dc: DiagnosticCategory; { to traverse first index of diagCount }
    b : Boolean;            { to traverse second index of diagCount }
  begin

    for dc := FirstDiagnostic to LastDiagnostic do begin
      for b := False to True do begin
        diagCount [ dc, b ] := 0
      end { for b }
    end { for dc }

  end; { InitDiagCount }

procedure WriteDiagCount ( var outFile: Text;
                           const diagCount: DiagnosticCount  );
  { pre:  outFile is open for writing,
          diagCount is initialized
    post: message counts in diagCount are written to outFile,
          filtered by theDiagSel, unless DebugInfo requested (then all) }

  procedure WriteDCLine ( rep, unrep: LongInt;
                          caption: String );
    { pre:  true
      post: line written of the form "<rep>  <caption>" if not DebugInfo,
            else of the form "<rep>  <caption>  ( <unrep>  not reported )" }
    begin
      if ( unrep <> 0 ) then begin

        while  Length ( caption ) < MaxDCStringLen do begin
          caption := caption + Blank
        end { while }

      ; caption := caption + ColSep + '(' +
                   IntToStr ( unrep, NumWidth ) + ColSep +
                   'not reported )'
      end { if }
    ; writeln ( outFile, Indent, rep : NumWidth, ColSep, caption )
    end; { WriteDCLine }

  var
    dc    : DiagnosticCategory; { to traverse diagCount }
    totRep: LongInt; { to calculate number of messages reported }
    totUnr: LongInt; { to calculate number of unreported messages (Debug) }
    i     : Integer; { to count off dashes }

  begin { WriteDiagCount}
  ; totRep := 0
  ; totUnr := 0

  ; for dc := FirstDiagnostic to LastDiagnostic do begin
      if ( ( diagCount [ dc, True ] > 0 ) ) or
         ( DebugInfo in theDiagSel ) then begin
        WriteDCLine ( diagCount [ dc, True ], diagCount [ dc, False ],
                      StringFromDC ( dc ) )
      ; Inc ( totRep, diagCount [ dc, True ] )
      ; Inc ( totUnr, diagCount [ dc, False ] )
      end { if }
    end { for dc }

    { write indented line with NumWidth dashes as vertical separator }
  ; write ( outFile, Indent )

  ; for i := 1 to NumWidth do begin
      write ( outFile, '-' )
    end { for i }

  ; writeln ( outFile )
  ; WriteDCLine ( totRep, totUnr, 'Messages reported' )
  ; writeln ( outFile )
  ; Flush ( outFile )
  end; { WriteDiagCount }

procedure InitMsgCount ( var msgCount: MessageCount;
                         value: MsgCountRange );
  { pre:  true
    post: msgCount [ m ] = value, for all m }
  var
    msgNr: MessageNumber; { to traverse msgCount }
  begin

    for msgNr := SpecialMessageNr to MaxMessageNumber do begin
      msgCount [ msgNr ] := value
    end { for msgNr }

  end; { InitMsgCount }

function StringFromCount ( count: LongInt;
                           lowerBound: Boolean ): String;
  { pre:  count > 0
    ret:  human readable string representation of count,
          if lowerBound then '>' ++ count - 1 }
  begin
    if lowerBound then begin
      StringFromCount := '>' + IntToStr ( count - 1, 1 )
    end { then }
    else begin
      StringFromCount := IntToStr ( count, 1 )
    end { else }
  end; { StringFromCount }

procedure WriteMsgCount ( var outFile: Text;
                          const msgCount: MessageCount );
  { pre:  outFile open for writing
    post: msgCount written to outFile, unless all non-positive }

  var
    msgNr: MessageNumber;    { to traverse msgCount }
    totSuppressed: LongInt;  { total number suppressed }
    lowerBound   : Boolean;  { totSuppressed is lower bound
                               rather than exact }
    countStr     : String;   { to hold string from count }
  begin
    totSuppressed := 0
  ; lowerBound    := False

  ; for msgNr := SpecialMessageNr to MaxMessageNumber do begin
      if msgCount [ msgNr ] > 0 then begin
        Inc ( totSuppressed, msgCount [ msgNr ] )
      ; lowerBound := lowerBound or ( msgCount [ msgNr ] >= MaxMsgCount )
      end { if }
    end { for msgNr }

  ; if totSuppressed > 0 then begin
      countStr := '' {StringFromCount ( totSuppressed, lowerBound )}
    ; writeln ( outFile, Indent,
                countStr : NumWidth, ColSep,
                'Messages beyond reporting limit (<count> x <message number>)' )

    ; write ( outFile, Indent, '' : NumWidth )

    ; for msgNr := SpecialMessageNr to MaxMessageNumber do begin
{ written on single line, could be very long, maybe should wrap }
        if msgCount [ msgNr ] > 0 then begin
          countStr := StringFromCount ( msgCount [ msgNr ],
                                        msgCount [ msgNr ] >= MaxMsgCount )
        ; write ( outFile, ColSep,
                  countStr : MaxMsgCntWidth, 'x',
                  msgNr : MaxMsgNrWidth )
        end { if }
      end { for msgNr }

    ; writeln ( outFile )

    ; writeln ( outFile )
    end { if }
  end; { WriteMsgCount }

procedure WriteDiagnosticHeader ( var outFile: Text;
                                  diagCat: DiagnosticCategory;
                                  msgNr: MessageNumber;
                                  const extraHeader, extraMsg: String );
  { pre:  outFile is open for writing
    post: diagnostic message header written to outFile }
  var
    msg: Message; { temporary to hold the message }
  begin
    write ( outFile, ModuleName, ': ',
            StringFromDC ( diagCat ), ' ', msgNr : MaxMsgNrWidth )
  ; if extraHeader <> '' then begin
      write ( outFile, ' ', extraHeader )
    end { if }
  ; writeln ( outFile )

  ; msg := GetMessage ( theMsgTbl, msgNr )
  ; write ( outFile, Indent, msg )
  ; if ( msg <> '' ) and ( extraMsg <> '' ) then begin
      write ( outFile, ' (' )
    end { if }
  ; if extraMsg <> '' then begin
      write ( outFile, extraMsg )
    end { if }
  ; if ( msg <> '' ) and ( extraMsg <> '' ) then begin
      write ( outFile, ')' )
    end { if }
  ; writeln ( outFile )
  end; { WriteDiagnosticHeader }

procedure FiniDiagnostic ( diagCat: DiagnosticCategory;
                           reported: Boolean;
                           abortDiagSet: DiagnosticSet;
                           msgNr: MessageNumber );
  { pre:  diagnostic has been reported
    post: theDiagCount updated according to reported,
          the log file flushed,
          if diagCat in abortDiagSet, then
            explanatory note written to standard output, and
            execution terminated }
  begin
    Inc ( theDiagCount [ diagCat, reported ] )
  ; Flush ( theLogFileP ^ )
  ; if diagCat in abortDiagSet then begin
      writeln ( StringFromDC ( diagCat ), ' ', msgNr, ' (see LOG file)' )
    ; Flush ( output )
    ; Halt ( 1 )
  end { if }
  end; { FiniDiagnostic }

procedure ReportDiagnostic ( const routineName: String;
                             diagCat: DiagnosticCategory;
                             msgNr: MessageNumber;
                             const extraMsg: String );
  { pre:  routineName <> ''
    post: diagnostic message processed, i.e. printed (if not filtered)
          and counted, execution terminated if appropriate
    usage note: for RIF-specific diagnostics, call ReportDiagnosticRIF }
  { impl note: ReportDiagnostic must not call routines that can cause
               calls to ReportDiagnostic (avoid mutual recursion) }
  { impl note: run-time error if writes to theLogFileP^ fail }
  begin
    if diagCat in theDiagSel then begin
      WriteDiagnosticHeader ( theLogFileP ^, diagCat, msgNr,
                              'in ' + routineName, extraMsg )
    end { if }

  ; FiniDiagnostic ( diagCat, diagCat in theDiagSel,
                     [ IOError, InternalError, UsageError ],
                     msgNr )
  end; { ReportDiagnostic }

procedure HandleIOResult ( const routineName: String;
                           const extraMsg: String );
  { pre:  an I/O operation has just been called with I/O checking off
    post: IOResult has been called, and reported if not zero }
  var
    ior: Integer;
  begin
    ior := IOResult
  ; if ior <> 0 then begin
      { I/O Error ior }
      ReportDiagnostic ( routineName, InternalError, ior, extraMsg  )
    end { if }
  end; { HandleIOResult }

procedure WriteMessageTable ( var outFile: Text;
                              const mt: MessageTable );
  { pre:  outFile is open for writing
    post: all message definitions in mt written to outFile }
  var
    m: MessageNumber; { to traverse message definitions }
  begin
    if mt = nil then begin
      writeln ( outFile, Indent, 'Message table does not exist' )
    end { then }
    else begin

      for m := SpecialMessageNr to MaxMessageNumber do begin
        with mt ^, tbl [ m ] do begin
          if ptr <> nil then begin
            writeln ( outFile, Indent,
                      m: MaxMsgNrWidth, ColSep,
                      '"', ptr ^, '"' )
          end { if }
        end { with }
      end { for m }

    end { else }
  end; { WriteMessageTable }

procedure InsertMsgDefinition ( var mt: MessageTable;
                                msgNr: MessageNumber;
                                const msg: Message;
                                var redefinedCount: LongInt );
  { pre:  mt <> nil
    post: msgNr is associated with msg in mt;
          if msgNr already defined,
            then it is redefined and redefinedCount is incremented }
  { impl note: run-time error if new fails }
  const
    RoutineName = 'InsertMsgDefinition';
  begin
    if mt = nil then begin
      { Message table does not exist }
      ReportDiagnostic ( RoutineName, InternalError, {M#}260, 'mt = nil' )
    end
  ; with mt ^, tbl [ msgNr ] do begin
      if ptr = nil then begin
        new ( ptr )
      end { then }
      else begin
        { Message redefined }
        ReportDiagnostic ( RoutineName, UsageWarning, {M#}370,
                           IntToStr ( msgNr, 1 ) )
      ; Inc ( redefinedCount )
      end { else }
    ; ptr ^ := msg
    ; if ( msgNr <> SpecialMessageNr ) and ( msg = '' ) then begin
        { Empty message }
        ReportDiagnostic ( RoutineName, UsageWarning, {M#}371,
                           IntToStr ( msgNr, 1 ) )
      end { if }
    end { with }
  end; { InsertMsgDefinition }

procedure LoadMsgDefinitions ( var mt: MessageTable;
                               mdfPath: String;
                               minMsgNr, maxMsgNr: Integer );
  { pre:  mt <> nil, mdfPath is path for message definition file
    post: mt has been extended with message definitions in the file;
          message numbers outside the range minMsgNr..maxMsgNr ignored;
          if a message number was already defined, then it is redefined }
  const
    RoutineName = 'LoadMsgDefinitions';
    CommentChar = '#';
  var
    mdfFile: Text;     { to read the message definition file }
    { WHAT IF FILE OPERATIONS ON msgFile FAIL? }
    lineNr: LongInt;   { line number of line read from msgFile }
    line  : String;    { line read from msgFile }
    msgNr : LongInt;   { message number extracted from line }
    msg   : String;    { message extracted from line }
    ior   : Integer;   { to capture IOResult }
    code  : Integer;   { to capture result code of Val }
    emptyLines,        { # empty lines read }
    commentLines,      { # comment lines read }
    messageLines,      { # valid message lines }
    redefinedCount,    { # messages redefined }
    badLines: LongInt; { # bad lines read }

  procedure ReadMdfLine;
    { pre:  not eof(mdfFile) (message definition file is not at end)
      post: one line of message defintion file is read and processed }
    const
      Ellipsis = ' ...'; { to mark truncated message }
    begin
      Inc ( lineNr )
      {$I-}
    ; readln ( mdfFile, line )
      {$I+}
    ; HandleIOResult ( RoutineName,
                       'Reading message file "' + mdfPath + '"' )
    ; line := TrimString ( line )
    ; if ( line = '' ) then begin
        { empty line, ignore }
        Inc ( emptyLines )
      end { then }
      else if ( line [ 1 ] = CommentChar ) then begin
        { comment line, ignore }
        Inc ( commentLines )
      end { then }
      else begin
        { non-empty, non-comment line,
          determine length of message number, if any }
        Val ( line, msgNr, code )
        { code is index of first character not part of a leading integer
          in line }
      ; if code = 0 then begin
          { string consists of message number only }
          msg := ''
        end { then }
        else begin
          { the first code-1 characters encode the message number }
          { the remaining Length(s)-code+1 characters are the message text }
          msg := TrimString ( Copy ( line, code, Length ( line ) - code + 1 ) )
          { get the message number, if any }
        ; Val ( Copy ( line, 1, code - 1 ), msgNr, code )
        end { else }
      ; if code = 0 then begin
          { message number is valid, check range }
          if ( msgNr < minMsgNr ) or ( msgNr > maxMsgNr ) then begin
            { Message number not in range }
            ReportDiagnostic ( RoutineName, UsageWarning, {M#}372,
                               IntToStr ( msgNr, 1) + ' not in [' +
                               IntToStr ( minMsgNr, 1 ) + '..' +
                               IntToStr ( maxMsgNr, 1 ) + ']' +
                               ', file ' + mdfPath +
                               ', line ' + IntToStr ( lineNr, 1 ) )
            { ignore line }
          ; Inc ( badLines )
          end { then }
          else begin
            { message number in range, check message length }
            if Length ( msg ) > MaxMessageLength then begin
              { Message too long }
              ReportDiagnostic ( RoutineName, UsageWarning, {M#}373,
                                 IntToStr ( msgNr, 1 ) +
                                 ', file ' + mdfPath +
                                 ', line ' + IntToStr ( lineNr, 1 ) )
              { truncate message }
            ; msg := Copy ( msg, 1,
                            MaxMessageLength - Length ( Ellipsis ) ) +
                     Ellipsis
            end { if }
          ; InsertMsgDefinition ( mt, msgNr, msg, redefinedCount )
          ; Inc ( messageLines )
          end { else }
        end { then }
        else { code <> 0 } begin
          { Invalid line in message definition file }
          ReportDiagnostic ( RoutineName, UsageWarning, {M#}374,
                             'file ' + mdfPath +
                             ', line ' + IntToStr ( lineNr, 1 ) )
          { ignore line }
        ; Inc ( badLines )
        end { else }
      end { else }
    end; { ReadMdfLine }

  procedure SummarizeMdf ( var outFile: Text );
    { pre:  outFile is open for writing
      post: various statistics about mdfFile written to outFile }
    var
      m     : MessageNumber; { to traverse mt^ }
      minNr,           { smallest message number }
      maxNr,           { greatest message number }
      msgLen,          { temporary to calculate message length }
      minLen,          { length of shortest message }
      maxLen: Integer; { length of longest message }
      tm    : Integer; { # messages present }
      tc    : LongInt; { # characters in messages }
    begin
      minNr := maxint
    ; maxNr := -maxint
    ; minLen := maxint
    ; maxLen := -maxint
    ; tm := 0
    ; tc := 0

    ; for m := minMsgNr to maxMsgNr do begin
        with mt ^, tbl [ m ] do begin
          if ptr <> nil then begin { message defined }
            msgLen := Length ( ptr ^ )
          ; if m < minNr then minNr := m
          ; if m > maxNr then maxNr := m
          ; if msgLen < minLen then minLen := msgLen
          ; if msgLen > maxLen then maxLen := msgLen
          ; Inc ( tm )
          ; Inc ( tc, msgLen )
          end { if }
        end { with }
      end { for m }

    ; writeln ( outFile, Indent,
                '# empty lines    = ', emptyLines     : NumWidth )
    ; writeln ( outFile, Indent,
                '# comment lines  = ', commentLines   : NumWidth )
    ; writeln ( outFile, Indent,
                '# bad lines      = ', badLines       : NumWidth )
    ; writeln ( outFile, Indent,
                '# messages       = ', tm             : NumWidth )
    ; writeln ( outFile, Indent,
                '# redefined      = ', redefinedCount : NumWidth )
    ; writeln ( outFile, Indent,
                '# lines total    = ', LineNr         : NumWidth )
    ; writeln ( outFile, Indent,
                '# msg characters = ', tc             : NumWidth )
    ; if tm <> 0 then begin
        writeln ( outFile, Indent,
                  'Smallest message number    = ', minNr  : MaxMsgNrWidth )
      ; writeln ( outFile, Indent,
                  'Greatest message number    = ', maxNr  : MaxMsgNrWidth )
      ; writeln ( outFile, Indent,
                  'Length of shortest message = ', minLen : MaxMsgNrWidth )
      ; writeln ( outFile, Indent,
                  'Length of longest message  = ', maxLen : MaxMsgNrWidth )
      end { if }
    end; { SummarizeMdf }

  begin { LoadMsgDefinitions}
    if mt = nil then begin
      { Message table does not exist }
      ReportDiagnostic ( RoutineName, InternalError, {M#}260, 'mt = nil' )
    end { if }
  ; if ( minMsgNr < MinMessageNumber ) or ( minMsgNr > maxMsgNr ) or
       ( maxMsgNr > MaxMessageNumber ) then begin
      { Invalid message number range }
      ReportDiagnostic ( RoutineName, InternalError, {M#}261,
                         IntToStr ( minMsgNr, 1 ) + '..' +
                         IntToStr ( maxMsgNr, 1 ) )
    end { if }
    {$I-}
  ; Assign ( mdfFile, mdfPath )
  ; reset ( mdfFile )
    {$I+}
  ; HandleIOResult ( RoutineName,
                     'Opening message file "' + mdfPath + '"' )
  ; lineNr         := 0
  ; emptyLines     := 0
  ; commentLines   := 0
  ; messageLines   := 0
  ; redefinedCount := 0
  ; badLines       := 0

  ; while not eof ( mdfFile ) do begin
      ReadMdfLine
    end { while }

    {$I-}
  ; Close ( mdfFile )
    {$I+}
  ; HandleIOResult ( RoutineName,
                     'Closing message file "' + mdfPath + '"' )
  ; ReportDiagnostic ( RoutineName, DebugInfo, {M#}891,
                       mdfPath + ', ' +
                       IntToStr ( minMsgNr, 1 ) + ', ' +
                       IntToStr ( maxMsgNr, 1 ) )
  ; if DebugInfo in theDiagSel then begin
      SummarizeMdf ( theLogFileP ^ )
    end { if }
  end; { LoadMsgDefinitions }

procedure NewMessageTable ( var mt: MessageTable );
  { pre:  true (preferably mt^ is not allocated to avoid memory leak)
    post: mt is newly allocated empty MessageTable }
  { impl note: run-time error if new fails }
  var
    msgNr: MessageNumber;
    dummy: LongInt; { to capture and ignore result from InsertMessage }
  begin
    new ( mt )
  ; with mt ^ do begin

      for msgNr := 0 to MaxMessageNumber do begin
        tbl [ msgNr ] . ptr := nil
      end { for msgNr }

    end { with mt ^ }
  ; dummy := 0
    { associate SpecialMessageNr with empty message }
  ; InsertMsgDefinition ( mt, SpecialMessageNr, '', dummy )
  end; { NewMessageTable }

procedure DisposeMessageTable ( var mt: MessageTable );
  { pre:  mt <> nil
    post: mt = nil and mt^ has been disposed }
  const
    RoutineName = 'DisposeMessageTable';
  var
    msgNr: MessageNumber;
  begin
    if mt = nil then begin
      { Message table does not exist }
      ReportDiagnostic ( RoutineName, InternalError, {M#}260, 'mt = nil' )
    end
  ; with mt ^ do begin

      for msgNr := 0 to MaxMessageNumber do begin
        with tbl [ msgNr ] do begin
          if ptr <> nil then begin
            Dispose ( ptr )
          end { if }
        end { with tbl ... }
      end { for m }

    end { with mt ^ }
  ; Dispose ( mt )
  ; mt := nil
  end; { DisposeMessageTable }

{ ----- End of ADT MessageTable ----- }


function  StringFromCL ( cl: CheckingLevel ): String;
  { pre:  true
    ret:  human readable string representation of cl }
  begin
    case cl of
      RelaxedChecking: StringFromCL := 'Relaxed Checking'
    ; StrictChecking : StringFromCL := 'Strict Checking'
    end { case cl }
  end; { StringFromCL }

function  DCFromCL ( cl: CheckingLevel ): DiagnosticCategory;
  { pre:  true
    ret:  DiagnosticCategory corresponding to cl }
  begin
    case cl of
      RelaxedChecking: DCFromCL := FormatWarning
    ; StrictChecking : DCFromCL := FormatError
    end { case cl }
  end; { DCFromCL }

function  StringFromICt ( ic: ItemCategory ): String;
  { pre:  true
    ret:  human readable string representation of ic }
  begin
    case ic of
      NoItem       : StringFromICt := '<none>'
    ; CharacterItem: StringFromICt := 'character'
    ; IntegerItem  : StringFromICt := 'integer'
    ; StringItem   : StringFromICt := 'string'
    end { case ic }
  end; { StringFromICt }

function  StringFromICp ( comp: ItemComposition ): String;
  { pre:  true
    ret:  human readable string representation of comp }
  begin
    case comp of
      Contiguous: StringFromICp := 'Contiguous'
    ; Separated : StringFromICp := 'Separated'
    end { case comp }
  end; { StringFromICp }

function  IsInitializedRIF ( const rif: RobInFile ): Boolean;
  { pre:  true
    ret:  whether rif is initialized
    usage note: Works correctly for rif <-- inFile (declared in RobIn).
                The current implementation may be slightly too optimistic
                for variables rif declared in the using program, that is,
                "rif is open" implies IsOpenRIF(rif), but possibly
                not the converse.  True may be returned when rif
                is in fact not open, though the probability is small. }
  { impl note: based on a signature }
  begin
    with rif do begin
      IsInitializedRIF := signature = theSignature
    end { with rif }
  end; { IsOpenRIF }

function  IsOpenRIF ( const rif: RobInFile ): Boolean;
  { pre:  true
    ret:  whether rif is open
    usage note: uses optimistic IsInitializedRIF }
  begin
    with rif do begin
      IsOpenRIF := IsInitializedRIF ( rif ) and rifOpen
    end { with rif }
  end; { IsOpenRIF }

function  FileCoordRIF ( const rif: RobInFile ): String;
  { pre:  rif is open
    ret:  human readable string representation of rif.nextChar
          and its coordinates within file rif.f }
  begin
    with rif do begin
      FileCoordRIF := 'File   ' + '"' + fPath + '"' +
                      ', line ' + IntToStr ( lineNr, NumWidth ) +
                      ', col. ' + IntToStr ( columnNr, NumWidth ) +
                      ', char ' + CharToStr ( nextChar )
    end { with rif }
  end; { FileCoordRIF }

function  FormatCoordRIF ( const rif: RobInFile ): String;
  { pre:  rif is open
    ret:  human readable string representation of current item
          and its coordinates within the format being processed
          through rif }
  var
    result: String; { to accumulate result }
  begin
    with rif do begin
      result := 'Format ' + '"' + fmtName + '"' +
                ', line ' + IntToStr ( fmtLineNr, NumWidth )  +
                ', item ' + IntToStr ( itemNr, NumWidth ) +
                ', ' + StringFromICt ( itemCat )
    ; if itemIdent <> '' then begin
        result := result + ' ' + itemIdent
      end { if }
    ; if itemValue <> '' then begin
        result := result + ' = ' + itemValue
      end { if }
    ; FormatCoordRIF := result
    end { with rif }
  end; { FormatCoordRIF }

procedure WriteStatisticsRIF ( var outFile: Text;
                               const rif: RobInFile );
  { pre:  outFile is open for writing, rif is open
    post: statistics for rif written to outFile }
  var
    lines : LongInt; { to calculate actual number of file lines read }
  begin
    with rif do begin
    ; if nextChar = EofChar then
        lines := lineNr - 1
      else
        lines := lineNr
    ; writeln ( outFile, Indent,
                lines : NumWidth, ColSep, 'lines with', ColSep,
            charCount : NumWidth, ColSep, 'chars read from file' )
    ; writeln ( outFile, Indent,
        fmtLineNr - 1 : NumWidth, ColSep, 'lines with', ColSep,
            itemCount : NumWidth, ColSep, 'items processed from format' )
    ; writeln ( outFile )
    ; WriteDiagCount ( outFile, diagCount )
    ; WriteMsgCount ( outFile, msgCount )
    end { with rif }
  end; { WriteStatisticsRIF }

procedure ReportDiagnosticRIF ( var rif: RobInFile;
                                diagCat: DiagnosticCategory;
                                msgNr: MessageNumber;
                                const extraMsg: String );
  { pre:  rif is open
    post: diagnostic message processed, i.e. printed (if not filtered)
          and counted, execution terminated if appropriate
  { impl note: run-time error if writes to theLogFileP^ fail }
  var
    report     : Boolean; { whether to report diagnostic }
    last       : Boolean; { whether last to report (only if report) }
    increment  : Boolean; { whether to increment message count }
    extraHeader: String; { to hold computed extra header }
  begin
    with rif do begin
      report    := ( diagCat in theDiagSel ) and
                   ( not ( diagCat in theLimitedDiag ) or
                     ( msgCount [ msgNr ] < 0 ) )
    ; last      := report and ( diagCat in theLimitedDiag ) and
                   ( msgCount [ msgNr ] = -1 )
    ; increment := ( diagCat in theLimitedDiag ) and
                   ( msgCount [ msgNr ] > - UnlimitedReporting ) and
                   ( msgCount [ msgNr ] < MaxMsgCount )

    ; if report then begin
        extraHeader := ''
      ; if last then begin
          extraHeader := '(not further reported)'
        end { if }
      ; WriteDiagnosticHeader ( theLogFileP ^, diagCat, msgNr,
                                extraHeader, extraMsg )

      ; if ( diagCat in [ UsageWarning .. AssertError,
                          ItemInfo, WhitespaceInfo ] ) or
           ( DebugInfo in theDiagSel ) then begin
          { write coordinates of item within format }
          writeln ( theLogFileP ^, Indent, FormatCoordRIF ( rif ) )
        end { if }

      ; if ( diagCat in [ InternalError,
                          FormatWarning, FormatError ] ) or
           ( ( diagCat = IOError ) and fOpened ) or
           ( DebugInfo in theDiagSel ) then begin
          { write coordinates of nextChar in file }
          writeln ( theLogFileP ^, Indent, FileCoordRIF ( rif ) )
        end { if }
      end { if }

    ; if diagCat in AbortReadDiagnostics then begin
        if not aborted and ( diagCat in theDiagSel ) then begin
          { report abortion (only once) }
          writeln ( theLogFileP ^, Indent, 'READING ABORTED' )
        end { if }
      ; aborted := True
        { Delay closing of f until CloseRIF,
          because closing may give rise to additional messages }
      end { if }

    ; Inc ( diagCount [ diagCat, report ] )
    ; Inc ( msgCount [ msgNr ], ord ( increment ) )
    ; FiniDiagnostic ( diagCat, report,
                       [ InternalError, UsageError ],
                       msgNr )
    end { with rif }
  end; { ReportDiagnosticRIF }

procedure HandleIOResultRIF ( var rif: RobInFile;
                              const extraMsg: String );
  { pre:  an I/O operation has just been called with I/O checking off
    post: IOResult has been called, and reported if not zero }
  var
    ior: Integer;
  begin
    ior := IOResult
  ; if ior <> 0 then begin
      { I/O Error ior }
      ReportDiagnosticRIF ( rif, IOError, ior, extraMsg )
    end { if }
  end; { HandleIOResultRIF }

procedure StepNextChar ( var rif: RobInFile );
  { pre:  true
    post: ( next character is assigned to nextChar,
            nextChar in [ EofChar, EolnChar ] + PrintableChars )
       \/ aborted }
  var
    eofTmp : Boolean; { auxiliary for checked eof call }
    eolnTmp: Boolean; { auxiliary for checked eoln call }
  begin
    with rif do begin
      if aborted then Exit

    ; if nextChar = EofChar then begin
        { Attempt to read past end-of-file }
        ReportDiagnosticRIF ( rif, InternalError, {M#}260, '' )
      ; Exit
      end { if }

    ; if nextChar = EolnChar then begin
        Inc ( lineNr )
      ; columnNr := 0 { will be incremented to 1 below,
                        unless eof or an error intervenes }
      end { if }

      { N.B. The order in which eof/eoln is called is important
        in Turbo Pascal.
        A file could be without final eoln, that is eof becomes
        true at same time as eoln becomes true.  We compensate for
        this by inserting EolnChar if eof holds and the line is nonempty
        (columnNr > 0). }
      {$I-}
    ; eofTmp := eof ( f )
      {$I+}
    ; HandleIOResultRIF ( rif, 'eof(f)' )
    ; if aborted then Exit
    ; if eofTmp then begin
        nextChar := EofChar
      ; if columnNr <> 0 then begin
          { compensate for missing end-of-line after last line }
          Inc ( columnNr )
          { Last line has no explicit end-of-line }
        ; ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}602, '' )
        ; nextChar := EolnChar
        end { if }
      end { then }
      else { not eof } begin
        Inc ( columnNr )
        {$I-}
      ; eolnTmp := eoln ( f )
        {$I+}
      ; HandleIOResultRIF ( rif, 'eoln(f)' )
      ; if aborted then Exit
      ; if eolnTmp then begin
          nextChar := EolnChar
          {$I-}
        ; readln ( f )
          {$I+}
        ; HandleIOResultRIF ( rif, 'readln(f)' )
        ; if aborted then Exit
        end { else if }
        else { not eoln } begin
          {$I-}
          read ( f, nextChar )
          {$I+}
        ; Inc ( charCount )
        ; HandleIOResultRIF ( rif, 'read(f,nextChar)' )
        ; if aborted then Exit
        ; if not ( nextChar in PrintableChars ) then begin
            if ( nextChar = TabChar ) and
               ( chLevel = RelaxedChecking ) then begin
              { Invalid character  }
              ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}610,
                                    'TAB --> '' ''' )
            ; nextChar := Blank
            end { then }
            else begin
              { Invalid character }
              ReportDiagnosticRIF ( rif, FormatError, {M#}610,
                                    CharToStr ( nextChar ) )
            end { else }
          end { if }
        ; if aborted then Exit
        end { else }
      end { else }
    { nextChar in [ EofChar, EolnChar ] + PrintableChars }
    end { with rif }
  end; { StepNextChar }

procedure HandleItemComposition ( var rif: RobInFile );
  { pre:  true
    post: not ( nextChar in [ EofChar, EolnChar, Blank ] ) \/ aborted }
  const
    ExpectedEolnCount = 0; { # eoln expected }
  var
    eolnCount : LongInt; { # eoln  skipped }
    blankCount: LongInt; { # blank skipped }
    expectedBlankCount: Integer; { # blank expected }
    diagCat   : DiagnosticCategory;
  begin
    with rif do begin
      { first always handle meta-format issues (also when aborted) }
      if aborted then Exit

    ; eolnCount := 0
    ; blankCount := 0
    ; case itemComp of
        Contiguous: begin
          expectedBlankCount := 0
        end { Contiguous }
      ; Separated: begin
          expectedBlankCount := 1
        end { Separated }
      end { case comp }
    ; diagCat := DCFromCL ( chLevel )

    ; while nextChar in Whitespace do begin
        case nextChar of
          EolnChar: begin
            if eolnCount = ExpectedEolnCount then begin
              { reported only once }
              if itemNr = 1 then begin
                { Empty line between two consecutive format lines }
                ReportDiagnosticRIF ( rif, diagCat, {M#}614, '' )
              end { then }
              else begin
                { End-of-line between items on same format line }
                ReportDiagnosticRIF ( rif, diagCat, {M#}611, '' )
              end { else }
            end { if }
          ; Inc ( eolnCount )
          end { EolnChar }
        ; Blank: begin
            if blankCount = expectedBlankCount then begin
              { Too many blanks before item (reported only once) }
              ReportDiagnosticRIF ( rif, diagCat, {M#}612, '' )
            end { if }
          ; Inc ( blankCount )
          end { Blank }
        end { case nextChar }
      ; StepNextChar ( rif )
      ; if aborted then Exit
      end { while }

    ; if ( itemComp = Separated ) and ( blankCount = 0 ) and
         ( eolnCount = 0 ) then begin
        { Blank missing between items }
        ReportDiagnosticRIF ( rif, diagCat, {M#}613, '' )
      end { if }
    ; if nextChar = EofChar then begin
        { Item missing }
        ReportDiagnosticRIF ( rif, FormatError, {M#}710, '' )
      end { if }
      { Composition with preceding item }
    ; ReportDiagnosticRIF ( rif, WhitespaceInfo, {M#}810,
                            StringFromICp ( itemComp ) +
                            ', ' + IntToStr ( blankCount, 1 ) + 'x blank' +
                            ', ' + IntToStr ( eolnCount,1 ) + 'x eoln' )
    end { with rif }
  end; { HandleItemComposition }

procedure InitReadItemRIF ( var rif: RobInFile;
                            pItemCat: ItemCategory;
                            comp: ItemComposition;
                            const itemId: String );
  { pre:  rif is open
    post: common initialization for Read<Item>RIF done }
  begin
    with rif do begin
      precItemCat := itemCat
    ; Inc ( itemNr )
    ; itemComp    := comp
    ; itemCat     := pItemCat
    ; itemIdent   := itemId
    ; itemValue   := ''

    ; if ItemNr = 1 then begin
        { first item on line is always contiguous to left margin }
        itemComp := Contiguous
      end { if }
    end { with RIF }
  end; { InitReadItemRIF }

procedure FiniReadItemRIF ( var rif: RobInFile;
                            pItemCat: ItemCategory );
  { handles common finalization part of Read<Item>RIF }
  { usage note: parameter pItemCat currently not used }
  begin
    with rif do begin
      Inc ( itemCount )
    ; ReportDiagnosticRIF ( rif, ItemInfo, {M#}820, itemValue )
    end { with RIF }
  end; { InitReadItemRIF }


{ more global variables }
var
  openedRIF: LongInt; { # RIF ever opened }
  closedRIF: LongInt; { # RIF ever closed }

function  NOpenRIF: LongInt;
  { pre:  openedRIF, closedRIF initialized by BeginLogging
    ret:  # currently open RIF }
  begin
    NOpenRIF := openedRIF - closedRIF
  end; { NOpenRIF }

{ ----- Definitions for routines in interface ----- }
{ Specifications are not repeated here (maintenance nightmare :-) }

procedure BeginLogging ( var logFile: Text;
                         const logPath: String;
                         const diagSelection: DiagnosticSet;
                         reportingLimit: LimitRange;
                         const mdfDir, mdfPath: String );
  const
    RoutineName = 'BeginLogging';
  begin
    theDiagSel  := diagSelection { must be done first, otherwise user
                                   can never get DebugInfo on this entry }
  ; ReportDiagnostic ( RoutineName, DebugInfo, {M#}890, '' )
  ; if DebugInfo in theDiagSel then begin
    ; writeln ( theLogFileP ^, Indent,
                'diagSelection  = ', StringFromDS ( diagSelection ) )
    ; writeln ( theLogFileP ^, Indent,
                'reportingLimit = ', reportingLimit )
    ; writeln ( theLogFileP ^, Indent,
                'mdfDir         = "', mdfDir, '"' )
    ; writeln ( theLogFileP ^, Indent,
                'mdfPath        = "', mdfPath, '"' )
    end { if }

    { We have a chicken and egg problem here:
      In what order to handle opening of log file and loading of message
      definitions?  If message definitions loaded before opening log
      file, then messages raised by loading (UsageWarnings, UsageErrors,
      DebugInfo) go to old log file.  If log file opened before loading
      message definitions, then messages raised by opening are undefined.

      In case logPath <> ='', I have chosen to load message definitions
      first, because normally that should not raise messages, and because
      normally the messages raised by opening the log file should
      be defined.  In case logPath = '', the order is reversed, because
      apparently the using program has opened the log file itself.
    }

  ; if logPath = '' then begin
    ; Flush ( theLogFileP ^ )
      { switch the log file }
    ; theLogFileP := @ logFile
    ; WriteIdentification ( theLogFileP ^ )
    end { if }

  ; if theMsgTbl = nil then begin
      NewMessageTable ( theMsgTbl )
    ; LoadMsgDefinitions ( theMsgTbl, mdfDir + 'TPRTerr.mdf',    1, 255 )
    ; LoadMsgDefinitions ( theMsgTbl, mdfDir + 'RobInGen.mdf', 256, 899 )
    end { if }

  ; if logPath <> '' then begin
      {$I-}
      Assign ( logFile, logPath )
    ; rewrite ( logFile )
      {$I+}
    ; HandleIOResult ( RoutineName, 'Opening logFile "' + logPath + '"' )
      { File opened for logging }
    ; ReportDiagnostic ( RoutineName, OpenCloseInfo, {M#}880,
                         '"' + logPath + '"' )
    ; Flush ( theLogFileP ^ )
      { switch the log file }
    ; theLogFileP := @ logFile
    ; WriteIdentification ( theLogFileP ^ )
    end { if }

  ; theReportingLimit := reportingLimit
  ; InitDiagCount ( theDiagCount )
  ; openedRIF := 0
  ; closedRIF := 0
  ; if mdfPath <> '' then begin
      LoadMsgDefinitions ( theMsgTbl, mdfPath,
                           MinAssertMsgNr, MaxAssertMsgNr )
    end { if }
  ; ReportDiagnostic ( RoutineName, DebugInfo, {M#}891, 'Message Table' )
  ; if DebugInfo in theDiagSel then begin
      WriteMessageTable ( theLogFileP ^, theMsgTbl )
    end { if }
  ; Flush ( theLogFileP ^ )
  end; { BeginLogging }

procedure EndLogging ( closeLogFile: Boolean );
  const
    RoutineName = 'EndLogging';
  begin
  ; ReportDiagnostic ( RoutineName, DebugInfo, {M#}890,
                       BoolToStr ( closeLogFile ) )
  ; if NOpenRIF <> 0 then begin
      { RIF still open }
      ReportDiagnostic ( RoutineName, UsageWarning, {M#}302,
                         IntToStr ( NOpenRIF, 1 ) + 'x' )
    end { if }

  ; ReportDiagnostic ( RoutineName, OpenCloseInfo, {M#}882, '' )
  ; if OpenCloseInfo in theDiagSel then begin
      writeln ( theLogFileP ^, Indent,
                openedRIF : NumWidth, ColSep, 'files processed' )
    ; writeln ( theLogFileP ^ )
    ; WriteDiagCount ( theLogFileP ^, theDiagCount )
    end { if }
  ; if closeLogFile then begin
      Close ( theLogFileP ^ )
    ; theLogFileP := @ output
      { File closed for logging }
    ; ReportDiagnostic ( RoutineName, OpenCloseInfo, {M#}881, '' )
    end { if }
  ; DisposeMessageTable ( theMsgTbl )
  end; { EndLogging }

procedure OpenRIF ( var rif: RobInFile;
                    const formatName, filePath: String;
                    chkLvl: CheckingLevel );
  const
    RoutineName = 'OpenRIF';
  begin
    ReportDiagnostic ( RoutineName, DebugInfo, {M#}890, '' )
  ; if DebugInfo in theDiagSel then begin
      writeln ( theLogFileP ^, Indent,
                'formatName = "', formatName, '"' )
    ; writeln ( theLogFileP ^, Indent,
                'filePath   = "', filePath, '"' )
    ; writeln ( theLogFileP ^, Indent,
                'chkLvl     = ', StringFromCL ( chkLvl ) )
    end { if }

  ; if formatName = '' then begin
      ReportDiagnostic ( RoutineName, UsageWarning, {M#}300, 'formatName' )
    end { if }
  ; if filePath = '' then begin
      ReportDiagnostic ( RoutineName, UsageWarning, {M#}300, 'filePath' )
    end { if }
  ; if IsOpenRIF ( rif ) then begin
      { RIF already open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}303,
                         '"' + rif.fPath + '", "' + filePath + '"' )
    end { if }
    { rif not yet open }

  ; with rif do begin
      fmtName     := UpCaseString ( formatName )
    ; fPath       := UpCaseString ( filePath )
    ; fOpened     := False { anticipating an error }
    ; chLevel     := chkLvl
    ; precItemCat := NoItem { not relevant }
    ; itemComp    := Contiguous { not relevant }
    ; itemCat     := NoItem
    ; itemIdent   := ''
    ; itemValue   := ''
    ; nextChar    := EolnChar { dummy, skipped by StepNextChar below,
                                unless open fails }
    ; lineNr      := 0 { not 1, because nextChar precedes actual content }
    ; columnNr    := 0
    ; charCount   := 0
    ; fmtLineNr   := 1
    ; itemNr      := 0
    ; itemCount   := 0
    ; aborted     := False
    ; InitDiagCount ( diagCount )
    ; InitMsgCount  ( msgCount, - theReportingLimit )
    ; rifOpen     := True
    ; signature   := theSignature

      { Opening file for robust reading }
    ; ReportDiagnosticRIF ( rif, OpenCloseInfo, {M#}800,
                            '"' + fPath + '"' )
    ; if OpenCloseInfo in theDiagSel then begin
        writeln ( theLogFileP ^, Indent,
                  'Format "' + fmtName + '", ' +
                  StringFromCL ( chLevel ) )
{WRITE REPORTING LIMIT...?}
      end { if }
{ messages above too complicated for competitors? }

      {$I-}
    ; Assign ( f, fPath )
    ; reset ( f )
      {$I+}
      { N.B. From Help info for IOResult:
        If an I/O error occurs and I/O-checking is off, all subsequent I/O
        operations are ignored until a call is made to IOResult.
        Hence, the following call to HandleIOResultRIF works for both
        Assign and reset. }
    ; HandleIOResultRIF ( rif, 'Opening file "' + fPath + '"' )
    ; StepNextChar ( rif ) { to prime nextChar }
    ; if aborted then begin
        { open f to a known-existing file, so that
          "rif is open" == "rif.f is open" }
        {$I-}
        Assign ( f, '' ) { to stdin }
      ; reset ( f )
        {$I+}
      ; HandleIOResult ( RoutineName, 'Opening to stdin' )
      ; if [ MetaFormatWarning, MetaFormatError ] * theDiagSel <> [ ]
        then begin
          { Only checking meta-format rules }
        ; ReportDiagnosticRIF ( rif, OpenCloseInfo, {M#}802,
                                '"' + fmtName + '"' )
        end { if }
      end { if }
      else begin
        fOpened := True
      end { else }
    end { with rif }

  ; Inc ( openedRIF )
  ; if NOpenRIF > 1 then begin
      { Multiple RIF open simultaneously }
      ReportDiagnostic ( RoutineName, UsageWarning, {M#}304,
                         IntToStr ( NOpenRIF, 1 ) + 'x' )
    end { if }
  end; { OpenRIF }

procedure CloseRIF ( var rif: RobInFile );
  const
    RoutineName = 'CloseRIF';

  procedure PreClose;
    { pre:  meta-format checks were done
      post: an eof has been read \/ aborted }
    { Exits immediately when aborted.
      This way, even when format errors occur, the meta-format parts,
      in particular the updates after ActualReadln, are always handled. }
    begin
      with rif do begin
        if aborted then Exit
      ; if nextChar in Whitespace then begin
          { Whitespace after last line }
          ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}600, '' )

        ; repeat
            StepNextChar ( rif )
          ; if aborted then Exit
          until not ( nextChar in Whitespace )

        end { if }
        { nextChar not in WhiteSpace }
      ; if nextChar <> EofChar then begin
          { Non-whitespace after last line }
          ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}601, '' )
        end { if }
      end { with rif }
    end; { PreClose }

  begin { CloseRIF }
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890, RoutineName )
  ; if not IsOpenRIF ( rif ) then begin
      { RIF not open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}301, '' )
    end { if }
    { rif is open }
  ; if NOpenRIF = 0 then begin
      { NOpenRIF count incorrect }
      ReportDiagnostic ( RoutineName, InternalError, {M#}271, '' )
    end { if }

  ; with rif do begin
      if ( itemCat <> NoItem ) then begin
        { ReadlnRIF for last line missing }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}401, '' )
      ; ReadlnRIF ( rif )
      end { if }
    ; itemIdent := '*eof'
    ; itemValue := ''
    ; if ( fmtLineNr = 1 ) and ( itemNr = 0 ) then begin
        { Empty file prescribed }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}500, '' )
      end { if }

    ; PreClose

      {$I-}
    ; Close ( f )
      {$I+}
    ; if fOpened then begin
        HandleIOResultRIF ( rif, 'Closing file "' + fPath + '"' )
      end { then }
      else begin
        HandleIOResultRIF ( rif, 'Closing file stdin = ""' )
      ; if [ MetaFormatWarning, MetaFormatError ] * theDiagSel <> [ ]
        then begin
          { Meta-format check ended }
        ; ReportDiagnosticRIF ( rif, OpenCloseInfo, {M#}803,
                              '"' + fmtName + '"' )
        end { if }
      end { else }

      { File closed for robust reading }
    ; ReportDiagnosticRIF ( rif, OpenCloseInfo, {M#}801,
                            '"' + fPath + '"' )
    ; rifOpen := False
    ; Inc ( closedRIF )

    ; if OpenCloseInfo in theDiagSel then begin
        writeln ( theLogFileP ^ )
      ; WriteStatisticsRIF ( theLogFileP ^, rif )
      ; Flush ( theLogFileP ^ )
      end { if }
    end { with }
  end; { CloseRIF }

function  ReadCharRIF ( var rif: RobInFile;
                        comp: ItemComposition;
                        const itemId: String;
                        range: CharSet ): Char;
  const
    RoutineName = 'ReadCharRIF';

  var
    c: Char; { auxiliary to construct return value }

  procedure ActualReadChar;
    { pre:  meta-format checks were done
      post: a character has been read and assigned to c \/ aborted }
    { Exits immediately when aborted.
      This way, even when format errors occur, the meta-format parts,
      in particular the updates after ActualReadInteger, are always done. }
    begin
      with rif do begin
        HandleItemComposition ( rif )
      ; if aborted then Exit

      ; if nextChar in range then begin
          c := nextChar
        end { then }
        else begin
          { Character not in range }
          ReportDiagnosticRIF ( rif, FormatError, {M#}720,
                                CharToStr ( nextChar ) )
        end { else }
      ; StepNextChar ( rif )
      end { with rif }
    end; { ActualReadChar }

  begin { ReadCharRIF }
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890,
                          RoutineName + ': ' + StringFromICp ( comp ) +
                          ', ' + itemId +
                          ', [' + StrCS ( range ) + ']' )
  ; if not IsOpenRIF ( rif ) then begin
      { RIF not open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}301, '' )
    end { if }
    { rif is open }
  ; with rif do begin
      InitReadItemRIF ( rif, CharacterItem, comp, itemId )

    ; if range = [ ] then begin
        { Empty character range }
        ReportDiagnosticRIF ( rif, UsageError, {M#}320, '' )
      end { if }
      { range not empty }

    ; c := MinCS ( range ) { minimum as default }
    ; ReadCharRIF := c     { anticipating an error }

    ; if SizeCS ( range ) = 1 then begin
        { Range allows only one item }
        ReportDiagnosticRIF ( rif, UsageWarning, {M#}321, CharToStr ( c ) )
      end { if }
    ; if precItemCat = IntegerItem then begin
        { Character follows integer on same line }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}420, '' )
      end { if }
    ; if comp = Separated then begin
        { Blank prescribed before character }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}421, '' )
      end { if }
    ; if ( precItemCat = IntegerItem ) and ( comp = Contiguous ) then begin
        { Character contiguous after integer }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}520, '' )
      end { if }
    ; if ( precItemCat = StringItem ) then begin
        { Preceding string not at end of line }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}541, '' )
      end { if }
    ; if not ( range <= PrintingChars ) then begin
        { Range contains invalid character }
{UsageError?}
        if Blank in range then
          ReportDiagnosticRIF ( rif, MetaFormatError, {M#}521,
                                CharToStr ( Blank ) )
        else
          ReportDiagnosticRIF ( rif, MetaFormatError, {M#}521, '' )
      end { if }

    ; ActualReadChar

    ; itemValue   := '''' + c + ''''
    ; ReadCharRIF := c
    ; FiniReadItemRIF ( rif, CharacterItem )
    end { with rif }
  end; { ReadCharRIF }

function  ReadIntegerRIF ( var rif: RobInFile;
                           comp: ItemComposition;
                           const itemId: String;
                           lowerBound, upperBound: LongInt ): LongInt;
  const
    RoutineName = 'ReadIntegerRIF';

  var
    i: LongInt; { auxiliary to construct return value }

  procedure ActualReadInteger;
    { pre:  meta-format checks were done
      post: an integer number has been read and assigned to i \/ aborted }
    { Exits immediately when aborted.
      This way, even when format errors occur, the meta-format parts,
      in particular the updates after ActualReadInteger, are always done. }
    var
      negative: Boolean;
      tmp: LongInt; { temporary for building up integer while reading it }
    begin
      with rif do begin
        HandleItemComposition ( rif )
      ; if aborted then Exit

      ; if nextChar = '-' then begin
          negative := True
        ; StepNextChar ( rif )
        ; if aborted then Exit
        end { then }
        else begin
          negative := False
        ; if nextChar = '+' then begin
            { Plus sign in integer }
            ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}630, '' )
          ; StepNextChar ( rif )
          ; if aborted then Exit
          end { if }
        end { else }

      ; if not ( nextChar in DigitChars ) then begin
          { Not an integer }
          ReportDiagnosticRIF ( rif, FormatError, {M#}730, '' )
        end { then }
        else { nextChar in Digits } begin
          tmp := 0

        ; repeat
            { N.B. In the following, negative and positive are NOT combined,
              for example, by first building up the value without sign and
              afterwards negating it if negative, because the overflow
              properties of positive and negative integers are slightly
              different.  In particular, negative numbers can be one
              bigger (in absolute value) then positive numbers.  Of course,
              it is not nice if negative values of - MaxLongInt - 1 are
              involved. }
            if negative then begin
              {$Q-}
              tmp := 10 * tmp - ord ( nextChar ) + ord ( '0' )
              {$Q+}
            ; if tmp > 0 then begin { arithmetic overflow occurred }
                { Integer overflows 32 bit }
                ReportDiagnosticRIF ( rif, FormatError, {M#}731, '' )
              ; Exit
              end { if }
            end { then }
            else { positive } begin
              {$Q-}
              tmp := 10 * tmp + ord ( nextChar ) - ord ( '0' )
              {$Q+}
            ; if tmp < 0 then begin { arithmetic overflow occurred }
                { Integer overflows 32 bit }
                ReportDiagnosticRIF ( rif, FormatError, {M#}731, '' )
              ; Exit
              end { if }
            end { else }
          ; StepNextChar ( rif )
          ; if aborted then Exit
          until not ( nextChar in DigitChars )

        ; if ( lowerBound <= tmp ) and ( tmp <= upperBound ) then begin
            i := tmp
          end { then }
          else begin
            { Integer not in range }
            ReportDiagnosticRIF ( rif, FormatError, {M#}732,
                                  IntToStr ( tmp, 1 ) + ' not in [' +
                                  IntToStr ( lowerBound, 1 ) + '..' +
                                  IntToStr ( upperBound, 1 ) + ']' )
          end { else }
        end { else }
      end { with rif }
    end; { ActualReadInteger }

  begin { ReadIntegerRIF }
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890,
                          RoutineName + ': ' + StringFromICp ( comp ) +
                          ', ' + itemId +
                          ', ' + IntToStr ( lowerBound, 1 ) +
                          ', ' + IntToStr ( upperBound, 1 ) )
  ; if not IsOpenRIF ( rif ) then begin
      { RIF not open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}301, '' )
    end { if }
    { rif is open }

  ; with rif do begin
      InitReadItemRIF ( rif, IntegerItem, comp, itemId )

    ; if lowerBound > upperBound then begin
        { Empty integer range }
        ReportDiagnosticRIF ( rif, UsageError, {M#}330,
                              IntToStr ( lowerBound, 1 ) + ' > ' +
                              IntToStr ( upperbound, 1 ) )
      end { if }

    ; i              := lowerBound { minimum value as default }
    ; ReadIntegerRIF := i          { anticipating an error }

    ; if lowerBound = upperBound then begin
        { Range allows only one item }
        ReportDiagnosticRIF ( rif, UsageWarning, {M#}322, IntToStr ( i, 1 ) )
      end { if }
    ; if precItemCat = CharacterItem then begin
        { Integer follows character on same line }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}430, '' )
      end { if }
    ; if comp = Contiguous then begin
        { No blank prescribed before integer }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}431, '' )
      end { if }
    ; if ( precItemCat = IntegerItem ) and ( comp = Contiguous ) then begin
        { Integer contiguous after integer }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}530, '' )
      end { if }
    ; if ( precItemCat = StringItem ) then begin
        { Preceding string not at end of line }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}541, '' )
      end { if }

    ; ActualReadInteger

    ; itemValue      := IntToStr ( i, 1 )
    ; ReadIntegerRIF := i
    ; FiniReadItemRIF ( rif, IntegerItem )
    end { with rif }
  end; { ReadIntegerRIF }

function  ReadStringRIF ( var rif: RobInFile;
                          comp: ItemComposition;
                          const itemId: String;
                          range: CharSet;
                          minLength, maxLength: Integer ): String;
  const
    RoutineName = 'ReadStringRIF';

  var
    s: String; { auxiliary to construct return value }

  procedure ActualReadString;
    { pre:  meta-format checks were done
      post: a string has been read and assigned to s \/ aborted }
    { Exits immediately when aborted.
      This way, even when format errors occur, the meta-format parts,
      in particular the updates after ActualReadString, are always done. }
    var
      tmp: String; { temporary for building up string while reading it }
    begin
      with rif do begin
        HandleItemComposition ( rif )
      ; if aborted then Exit

      ; tmp := ''

      ; while not ( nextChar in [ EofChar, EolnChar ] ) do begin
          { append a character to tmp }
          if Length ( tmp ) = maxLength then begin
            { String too long }
            ReportDiagnosticRIF ( rif, FormatError, {M#}741,
                                  'length > ' + IntToStr ( maxLength , 1 ) )
          ; Exit
          end { then }
          else { Length ( tmp ) < maxLength } begin
            if not ( nextChar in range ) then begin
              { String character not in range }
              ReportDiagnosticRIF ( rif, FormatError, {M#}740, '' )
            ; Exit
            end { if }
          ; tmp := tmp + nextChar
          end { else }
        ; StepNextChar ( rif )
        ; if aborted then Exit
        end { while }

      ; if minLength <= Length ( tmp ) then begin
{ HOW TO FIND THE BUG IN THE LINE ABOVE WHEN REPLACING tmp -> s ? }
          s := tmp
        end { then }
        else { Length ( tmp ) < minLength } begin
          { String too short }
          ReportDiagnosticRIF ( rif, FormatError, {M#}742,
                                IntToStr ( Length ( tmp ), 1 ) + ' < ' +
                                IntToStr ( minLength, 1 ) )
        end { else }
      end { with rif }
    end; { ActualReadString }

  begin { ReadStringRIF }
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890,
                          RoutineName + ': ' + StringFromICp ( comp ) +
                          ', ' + itemId +
                          ', [' + StrCS ( range ) + ']' +
                          ', ' + IntToStr ( minLength, 1 ) +
                          ', ' + IntToStr ( maxLength, 1 ) )
  ; if not IsOpenRIF ( rif ) then begin
      { RIF not open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}301, '' )
    end { if }
    { rif is open }

  ; with rif do begin
      InitReadItemRIF ( rif, StringItem, comp, itemId )

    ; if range = [ ] then begin
        { Empty character range }
        ReportDiagnosticRIF ( rif, UsageError, {M#}320, '' )
      end { if }
    ; if minLength < 0 then begin
        { Negative minimum string length }
        ReportDiagnosticRIF ( rif, UsageError, {M#}343,
                              IntToStr ( minLength, 1 ) )
      end { if }
    ; if minLength > maxLength then begin
        { Empty range for string length }
        ReportDiagnosticRIF ( rif, UsageError, {M#}341,
                              IntToStr ( minLength, 1 ) + ' > ' +
                              IntToStr ( maxLength, 1 ) )
      end { if }
    ; if maxLength > MaxStringLength then begin
        { Maximum string length exceeds 255 }
        ReportDiagnosticRIF ( rif, UsageError, {M#}342, '' )
      end { if }

    ; s := MinCS ( range )

    ; while Length ( s ) <> minLength do begin
        s := s + s [ 1 ]
      end { while }      { using minimum value as default }

    ; ReadStringRIF := s { anticipating an error }

    ; if ( minLength = maxLength ) and ( SizeCS ( range ) = 1 ) then begin
        { Range allows only one item }
        ReportDiagnosticRIF ( rif, UsageWarning, {M#}321, '"' + s + '"' )
      end { if }
    ; if ( minLength = 1 ) and ( maxLength = 1 ) then begin
        { String length equal one prescribed }
        ReportDiagnosticRIF ( rif, UsageWarning, {M#}340, '' )
      end { if }
{no warning for minLength = maxLength in general }
    ; if precItemCat <> NoItem then begin
        { String not alone on line }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}440, '' )
      end { if }
    ; if comp = Separated then begin
        { Blank prescribed before string }
        ReportDiagnosticRIF ( rif, MetaFormatWarning, {M#}441, '' )
      end { if }
    ; if ( precItemCat = IntegerItem ) and ( comp = Contiguous ) then begin
        { String contiguous after integer }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}540, '' )
      end { if }
    ; if ( precItemCat = StringItem ) then begin
        { Preceding string not at end of line }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}541, '' )
      end { if }
    ; if not ( range <= PrintingChars ) then begin
        { Range contains invalid character }
{UsageError?}
        if Blank in range then
          ReportDiagnosticRIF ( rif, MetaFormatError, {M#}521, 'blank' )
        else
          ReportDiagnosticRIF ( rif, MetaFormatError, {M#}521, '' )
      end { if }
    ; if minLength < 1 then begin
        { Empty string allowed }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}542, '' )
      end { if }

    ; ActualReadString

    ; itemValue     := '"' + s + '"'
    ; ReadStringRIF := s
    ; FiniReadItemRIF ( rif, StringItem )
    end { with rif }
  end; { ReadStringRIF }

procedure ReadlnRIF ( var rif: RobInFile );
  const
    RoutineName = 'ReadlnRIF';

  var
    blankCount: LongInt; { # blank skipped }

  procedure ActualReadln;
    { pre:  meta-format checks were done
      post: an eoln has been read \/ aborted }
    { Exits immediately when aborted.
      This way, even when format errors occur, the meta-format parts,
      in particular the updates after ActualReadln, are always handled. }
    begin
      with rif do begin
        if aborted then Exit
      ; if nextChar = Blank then begin
          { Whitespace after last item on line }
          ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}660, '' )

        ; repeat
            StepNextChar ( rif )
          ; Inc ( blankCount )
          ; if aborted then Exit
          until nextChar <> Blank

        end { if }
        { nextChar <> Blank }
      ; if nextChar = EolnChar then begin
          StepNextChar ( rif )
        end { then }
        else { nextChar not in Whitespace } begin
          { N.B. nextChar <> EofChar, because of "compensation" in StepNextChar }
          { Non-whitespace after last item on line }
          ReportDiagnosticRIF ( rif, DCFromCL ( chLevel ), {M#}661, '' )
          { Two ways to handle this:
              1.  assume missing end-of-line (moving the junk to next line)
              2.  assume stray item (ignoring the junk)
            It is hard to guess "right".  We take the first approach,
            because it seems to be the fairest way (not ignoring any
            of the content, but giving it a chance on the next line).
          }
        end { else }
      end { with rif }
    end; { ActualReadln }

  begin { ReadlnRIF }
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890, RoutineName )
  ; if not IsOpenRIF ( rif ) then begin
      { RIF not open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}301, '' )
    end { if }
    { rif is open }

  ; with rif do begin
      precItemCat := itemCat
    ; Inc ( itemNr )
    ; itemComp := Contiguous { not relevant }
    ; itemCat := NoItem
    ; itemIdent := '*eoln'
    ; itemValue := ''

    ; if precItemCat = NoItem then begin
        { Empty line prescribed }
        ReportDiagnosticRIF ( rif, MetaFormatError, {M#}560, '' )
      end { if }

    ; blankCount := 0
    ; ActualReadln

      { End-of-line }
    ; ReportDiagnosticRIF ( rif, WhitespaceInfo, {M#}860,
                            IntToStr ( blankCount, 1 ) + 'x blank' )
    ; Inc ( fmtLineNr )
    ; itemNr := 0
    end { with rif }
  end; { ReadlnRIF }

procedure AssertRIF ( var rif: RobInFile;
                      condition: Boolean;
                      assertDiag: AssertDiagnostic;
                      msgNr: Integer;
                      const extraMsg: String );
  const
    RoutineName = 'AssertRIF';
  begin
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890, RoutineName +
                          ': ' + BoolToStr ( condition ) +
                          ', ' + IntToStr ( msgNr, 1 ) +
                          ', ' + extraMsg )
  ; if not IsOpenRIF ( rif ) then begin
      { RIF not open }
      ReportDiagnostic ( RoutineName, UsageError, {M#}301, '' )
    end { if }
    { rif is open }

  ; with rif do begin
    ; if not ( assertDiag in [ AssertWarning, AssertError ] ) then begin
        { Parameter assertDiag not in range }
        ReportDiagnosticRIF ( rif, UsageError, {M#}375,
                              StringFromDC ( assertDiag ) )
      end { if }
    ; if ( msgNr < MinAssertMsgNr ) or ( msgNr > MaxAssertMsgNr ) then begin
        { Message number not in range }
        ReportDiagnosticRIF ( rif, UsageError, {M#}372,
                              IntToStr ( msgNr, 1) + ' not in [' +
                              IntToStr ( MinMessageNumber, 1 ) + '..' +
                              IntToStr ( MaxMessageNumber, 1 ) + ']' )
      end { if }
    ; if aborted then Exit

    ; if not condition then begin
        { Task-specific diagnostic }
        ReportDiagnosticRIF ( rif, assertDiag, msgNr, extraMsg )
      end { if }
    end { with rif }
  end; { AssertRIF }

function  IsAbortedRIF ( var rif: RobInFile ): Boolean;
  { impl note: uses slightly too optimistic IsInitializedRIF }
  const
    RoutineName = 'IsAbortedRIF';
  begin
    ReportDiagnosticRIF ( rif, DebugInfo, {M#}890, RoutineName )
  ; with rif do begin
    ; if not IsInitializedRIF ( rif ) then begin
        { RIF never opened }
        ReportDiagnostic ( RoutineName, UsageError, {M#}305, '' )
      end { if }
    ; IsAbortedRIF := aborted
    end { with rif }
  end; { IsAbortedRIF }

function  IntToStr ( i: LongInt;
                     width: integer ): String;
  var
    result: String;
  begin
    Str ( i : width, result )
  ; IntToStr := result
  end; { IntToStr }

begin { RobIn initialization }

{$IFDEF TEST}
  WriteIdentification ( output )
{$ENDIF}
; InitSignature ( theSignature )

; theMsgTbl         := nil
; theLogFileP       := @ output
; theDiagSel        := OrganizerDiagnostics
; theReportingLimit := UnlimitedReporting
  { preceding can be overridden by BeginLogging }
; theLimitedDiag    := [ UsageWarning, MetaFormatWarning,
                         FormatWarning, AssertWarning ];
  { override preceding only if you know exactly what you are doing! }

; with inFile do begin
    aborted    := True  { not needed, just a precaution }
  ; rifOpen    := False { not needed, just a precaution }
  ; signature  := theSignature - 1 { NOT yet initialized by OpenRIF }
    { why - 1 and not + 1 or so?  That way previous runs of this program
      will not leave behind memory patterns that later can be interpreted
      as initialized.  Note that initialization is derived from the
      clock and later runs produce greater theSignature values }
  end { with inFile }

end.